package damp.ekeko;
import java.util.Collection;
import java.util.concurrent.ConcurrentHashMap;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IProjectDescription;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceDelta;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.ITypeHierarchy;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.CompilationUnit;

public class EkekoModel {
	
	public static void toggleNature(IProject project) {
		try {
			IProjectDescription description = project.getDescription();
			String[] natures = description.getNatureIds();

			for (int i = 0; i < natures.length; ++i) {
				if (EkekoNature.NATURE_ID.equals(natures[i])) {
					// Remove the nature
					String[] newNatures = new String[natures.length - 1];
					System.arraycopy(natures, 0, newNatures, 0, i);
					System.arraycopy(natures, i + 1, newNatures, i,
							natures.length - i - 1);
					description.setNatureIds(newNatures);
					project.setDescription(description, null);
					return;
				}
			}

			// Add the nature
			String[] newNatures = new String[natures.length + 1];
			System.arraycopy(natures, 0, newNatures, 0, natures.length);
			newNatures[natures.length] = EkekoNature.NATURE_ID;
			description.setNatureIds(newNatures);
			project.setDescription(description, null);
		} catch (CoreException e) {
		}
	}
	
	
	
	private EkekoModel() {
		clean();
	}
	
	// see http://en.wikipedia.org/wiki/Initialization_on_demand_holder_idiom
    private static class LazyHolder {
        public static final EkekoModel INSTANCE = new EkekoModel();
    }

    public static EkekoModel getInstance() {
		return LazyHolder.INSTANCE;
	}
		
	//to be manipulated by builder, nature and changedeltavisitor
	private java.util.concurrent.ConcurrentHashMap<IProject,ProjectModel> projectModels;	
		
	public ProjectModel getProjectModel(IProject ip) {
		return projectModels.get(ip);
	}
	
	public boolean hasProjectModel(IProject ip) {
		return getProjectModel(ip) != null;
	}
	
	public void removeProjectModel(IProject ip) {
		System.out.println("Removing an existing project from the model: " + ip.toString());
		projectModels.remove(ip);
	}
	
	public ProjectModel getProjectModel(IResource rsc) {
		return getProjectModel(rsc.getProject());
	}
	
	public JavaProjectModel getJavaProjectModel(IJavaProject p) {
		IResource rsc = p.getResource();
		if(rsc == null)
			return null;
		return (JavaProjectModel) getProjectModel(rsc);
	}
	
	public Collection<ProjectModel> getProjectModels() {
		return projectModels.values();
	}

	public ITypeHierarchy getTypeHierarchy(IType type) throws JavaModelException {
		IJavaProject jp = type.getJavaProject();
		if(jp == null)
			return null;
		JavaProjectModel jpm = getJavaProjectModel(jp);
		if(jpm == null)
			return null;
		return jpm.getTypeHierarchy(type);
	}

	
	public void clean() {
		System.out.println("Cleaning the model.");
		projectModels = new ConcurrentHashMap<IProject, ProjectModel>();
	}
	
	public void populate() {
		System.out.println("Populating model from enabled natures.");
		IProject[] iprojects = ResourcesPlugin.getWorkspace().getRoot().getProjects();
		for(IProject p : iprojects) {
			try {
				if(p.isOpen() && p.hasNature(EkekoNature.NATURE_ID)) 
					fullProjectBuild(p,null);
			} catch (CoreException e) {
				e.printStackTrace();
			}
		}
	}
	
	private ProjectModel newProjectModel(IProject p) throws CoreException {
		if(p.isNatureEnabled(JavaCore.NATURE_ID)) 
			return new JavaProjectModel(p);
		return new ProjectModel(p);
	}
	
	
	public void fullProjectBuild(IProject project, IProgressMonitor monitor) throws CoreException {
		if(project.isOpen()) {
			System.out.println("Adding a new project to the model: " + project.toString());
			ProjectModel model = newProjectModel(project);
			model.populate(monitor);
			projectModels.put(project,model);
		}
	}

	public void incrementalProjectBuild(IResourceDelta delta, IProgressMonitor monitor) throws CoreException {
		IResource resource = delta.getResource();
		int resourceType = resource.getType();
		IProject ip = resource.getProject();	
		if(resourceType == IResource.PROJECT) {
			if (!ip.hasNature(EkekoNature.NATURE_ID)) {
				System.out.println("Removing project from model as nature was removed: " + ip.toString());
				removeProjectModel(ip);
				return;
			}					
			switch (delta.getKind()) {
			case IResourceDelta.ADDED:	
				fullProjectBuild(ip, monitor);
				return;
			case IResourceDelta.REMOVED:
				removeProjectModel(ip);
				return;
			}
		}
		System.out.println("Updating a project in the model: " + ip.toString());
		getProjectModel(ip).processDelta(delta,monitor);		
	}
}

	
	
